"""
Cloning resources.

"""
from marshmallow import (
    Schema,
    fields,
    post_load,
    pre_dump,
)


class EdgeSchema(Schema):
    """
    An edge between UUID node ids.

    """
    fromId = fields.UUID(
        attribute="from_id",
        required=True,
    )
    toId = fields.UUID(
        attribute="to_id",
        required=True,
    )


class SubstitutionSchema(Schema):
    """
    A substitution from one UUID id to another.

    This schema is identical to an Edge currently, but is kept distinct in order
    to support non-UUID substitutions if ever needed.

    """
    fromId = fields.UUID(
        attribute="from_id",
        required=True,
    )
    toId = fields.UUID(
        attribute="to_id",
        required=True,
    )


class DAGSchema(Schema):
    """
    Represents a DAG.

    Nodes should be overridden with a non-raw schema.

    """
    nodes = fields.Nested(
        fields.Raw,
        required=True,
        attribute="nodes_map",
    )
    edges = fields.List(
        fields.Nested(EdgeSchema),
        required=True,
    )
    substitutions = fields.List(
        fields.Nested(SubstitutionSchema),
        missing=[],
        required=False,
    )

    @pre_dump
    def unflatten(self, obj, **kwargs):
        """
        Translate substitutions dictionary into objects.

        """
        obj.substitutions = [
            dict(from_id=key, to_id=value)
            for key, value in getattr(obj, "substitutions", {}).items()
        ]
        return obj


class NewCloneSchema(Schema):
    commit = fields.Boolean(
        missing=True,
        required=False,
    )
    substitutions = fields.List(
        fields.Nested(SubstitutionSchema),
        missing=[],
        required=False,
    )

    @post_load
    def flatten(self, obj, **kwargs):
        """
        Translate substitutions into a dictionary.

        """
        obj["substitutions"] = {
            item["from_id"]: item["to_id"]
            for item in obj["substitutions"]
        }
        return obj


class DAGCloningController:

    def __init__(self, store):
        self.store = store

    def explain(self, **kwargs):
        """
        Return a DAG "explaining" what may be cloned.

        """
        return self.store.explain(**kwargs)

    def clone(self, substitutions, commit=True, **kwargs):
        """
        Clone a DAG, optionally skipping the commit.

        """
        return self.store.clone(substitutions, **kwargs)



"""
Generates a Swagger definition for registered endpoints.

Note that:
 -  Swagger operations and type names use different conventions from the internal definitions
    because we want to make usage friendly for code generation (e.g. bravado)

 -  Marshmallow to JSON Schema conversion is somewhat simplistic. There are several projects
    that already implement this conversion that we could try adapting. At the moment, the
    overhead of adapting another library's conventions is too high.

 -  All resource definitions are assumed to be shared and are declared in the "definitions"
    section of the result.

 -  All errors are defined generically.


"""
from logging import getLogger
from urllib.parse import unquote

from openapi import model as swagger
from werkzeug.routing import BuildError

from microcosm_flask.conventions.registry import (
    get_qs_schema,
    get_request_schema,
    get_response_schema,
)
from microcosm_flask.errors import ErrorContextSchema, ErrorSchema, SubErrorSchema
from microcosm_flask.naming import name_for
from microcosm_flask.operations import Operation
from microcosm_flask.swagger.api import build_parameter, iter_schemas
from microcosm_flask.swagger.naming import operation_name, type_name


logger = getLogger("microcosm_flask.swagger")


def build_swagger(graph, ns, operations):
    """
    Build out the top-level swagger definition.

    """
    base_path = graph.build_route_path(ns.path, ns.prefix)
    schema = swagger.Swagger(
        swagger="2.0",
        info=swagger.Info(
            title=graph.metadata.name,
            version=ns.version,
        ),
        consumes=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        produces=swagger.MediaTypeList([
            swagger.MimeType("application/json"),
        ]),
        basePath=base_path,
        paths=swagger.Paths(),
        definitions=swagger.Definitions(),
    )
    add_paths(schema.paths, base_path, operations)
    add_definitions(schema.definitions, operations)
    try:
        schema.validate()
    except Exception:
        logger.exception("Swagger definition did not validate against swagger schema")
        raise

    return schema


def add_paths(paths, base_path, operations):
    """
    Add paths to swagger.

    """
    for operation, ns, rule, func in operations:
        path = build_path(operation, ns)
        if not path.startswith(base_path):
            continue
        method = operation.value.method.lower()
        # If there is no version number or prefix, we'd expect the base path to be ""
        # However, OpenAPI requires the minimal base path to be "/"
        # This means we need branching logic for that special case
        suffix_start = 0 if len(base_path) == 1 else len(base_path)
        paths.setdefault(
            path[suffix_start:],
            swagger.PathItem(),
        )[method] = build_operation(operation, ns, rule, func)


def add_definitions(definitions, operations):
    """
    Add definitions to swagger.

    """
    for definition_schema in iter_definitions(definitions, operations):
        if definition_schema is None:
            continue
        if isinstance(definition_schema, str):
            continue

        for name, schema in iter_schemas(definition_schema):
            definitions.setdefault(name, swagger.Schema(schema))


def iter_definitions(definitions, operations):
    """
    Generate definitions to be converted to swagger schema.

    """
    # general error schema per errors.py
    for error_schema_class in [ErrorSchema, ErrorContextSchema, SubErrorSchema]:
        yield error_schema_class()

    # add all request and response schemas
    for operation, obj, rule, func in operations:
        yield get_request_schema(func)
        yield get_response_schema(func)


def build_path(operation, ns):
    """
    Build a path URI for an operation.

    """
    try:
        return ns.url_for(operation, _external=False)
    except BuildError as error:
        # we are missing some URI path parameters
        uri_templates = {
            argument: "{{{}}}".format(argument)
            for argument in error.suggested.arguments
        }
        # flask will sometimes try to quote '{' and '}' characters
        return unquote(ns.url_for(operation, _external=False, **uri_templates))


def body_param(schema):
    return swagger.BodyParameter(**{
        "name": "body",
        "in": "body",
        "schema": swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(schema))),
        }),
    })


def header_param(name, required=False, param_type="string"):
    """
    Build a header parameter definition.

    """
    return swagger.HeaderParameterSubSchema(**{
        "name": name,
        "in": "header",
        "required": required,
        "type": param_type,
    })


def query_param(name, field, required=False):
    """
    Build a query parameter definition.

    """
    parameter = build_parameter(field)
    parameter["name"] = name
    parameter["in"] = "query"
    parameter["required"] = False

    return swagger.QueryParameterSubSchema(**parameter)


def path_param(name, ns):
    """
    Build a path parameter definition.

    """
    if ns.identifier_type == "uuid":
        param_type = "string"
        param_format = "uuid"
    else:
        param_type = "string"
        param_format = None

    kwargs = {
        "name": name,
        "in": "path",
        "required": True,
        "type": param_type,
    }
    if param_format:
        kwargs["format"] = param_format
    return swagger.PathParameterSubSchema(**kwargs)


def build_operation(operation, ns, rule, func):
    """
    Build an operation definition.

    """
    swagger_operation = swagger.Operation(
        operationId=operation_name(operation, ns),
        parameters=swagger.ParametersList([
        ]),
        responses=swagger.Responses(),
        tags=[ns.subject_name],
    )

    # custom header parameter
    swagger_operation.parameters.append(
        header_param("X-Response-Skip-Null")
    )

    # path parameters
    swagger_operation.parameters.extend([
        path_param(argument, ns)
        for argument in rule.arguments
    ])

    # query string parameters
    qs_schema = get_qs_schema(func)
    if qs_schema:
        swagger_operation.parameters.extend([
            query_param(name, field)
            for name, field in qs_schema.fields.items()
        ])

    # body parameter
    request_schema = get_request_schema(func)
    if request_schema:
        swagger_operation.parameters.append(
            body_param(request_schema)
        )

    # sort parameters for predictable output
    swagger_operation.parameters.sort(key=lambda parameter: parameter["name"])

    add_responses(swagger_operation, operation, ns, func)
    return swagger_operation


def add_responses(swagger_operation, operation, ns, func):
    """
    Add responses to an operation.

    """
    # default error
    swagger_operation.responses["default"] = build_response(
        description="An error occurred",
        resource=type_name(name_for(ErrorSchema())),
    )

    if getattr(func, "__doc__", None):
        description = func.__doc__.strip().splitlines()[0]
    else:
        description = "{} {}".format(operation.value.name, ns.subject_name)

    if operation in (Operation.Upload, Operation.UploadFor):
        swagger_operation.consumes = [
            "multipart/form-data"
        ]

    # resource request
    request_resource = get_request_schema(func)
    if isinstance(request_resource, str):
        if not hasattr(swagger_operation, "consumes"):
            swagger_operation.consumes = []
        swagger_operation.consumes.append(request_resource)

    # resources response
    response_resource = get_response_schema(func)
    if isinstance(response_resource, str):
        if not hasattr(swagger_operation, "produces"):
            swagger_operation.produces = []
        swagger_operation.produces.append(response_resource)
    elif not response_resource:
        response_code = (
            204
            if operation.value.default_code == 200
            else operation.value.default_code
        )
        swagger_operation.responses[str(response_code)] = build_response(
            description=description,
        )
    else:
        swagger_operation.responses[str(operation.value.default_code)] = build_response(
            description=description,
            resource=response_resource,
        )


def build_response(description, resource=None):
    """
    Build a response definition.

    """
    response = swagger.Response(
        description=description,
    )
    if resource is not None:
        response.schema = swagger.JsonReference({
            "$ref": "#/definitions/{}".format(type_name(name_for(resource))),
        })
    return response



"""
Config sharing convention.

Reports service config protected by basic auth for securely running services
locally with realistic config.

"""
from json import dumps, loads

from microcosm.loaders.compose import PartitioningLoader

from microcosm_flask.audit import skip_logging
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import make_response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


class Config:
    """
    Wrapper around service config state.

    """
    def __init__(self, graph, include_build_info=True):
        self.graph = graph
        self.name = graph.metadata.name

    def to_dict(self):
        """
        Encode the name, the status of all checks, and the current overall status.

        """
        if not isinstance(self.graph.loader, PartitioningLoader):
            return dict(msg="Config sharing disabled for non-partioned loader")

        if not hasattr(self.graph.loader, "secrets"):
            return dict(msg="Config sharing disabled if no secrets are labelled")

        def remove_nulls(dct):
            return {key: value for key, value in dct.items() if value is not None}

        return loads(
            dumps(self.graph.loader.config, skipkeys=True, default=lambda obj: None),
            object_hook=remove_nulls,
        )


class ConfigDiscoveryConvention(Convention):

    def __init__(self, graph):
        super(ConfigDiscoveryConvention, self).__init__(graph)
        self.config_discovery = Config(graph)

    def configure_retrieve(self, ns, definition):
        @self.add_route(ns.singleton_path, Operation.Retrieve, ns)
        @skip_logging
        def current_config_discovery():
            response_data = self.config_discovery.to_dict()
            return make_response(response_data, status_code=200)


def configure_config(graph):
    """
    Configure the health endpoint.

    :returns: the current service configuration
    """
    ns = Namespace(
        subject=Config,
    )

    convention = ConfigDiscoveryConvention(
        graph,
    )
    convention.configure(ns, retrieve=tuple())
    return convention.config_discovery



"""
Convention base class.

"""
from werkzeug.exceptions import NotAcceptable

from microcosm_flask.conventions.encoding import find_response_format
from microcosm_flask.operations import Operation


def identity(x):
    return x


class RouteAlreadyRegisteredException(Exception):
    pass


class EndpointDefinition(tuple):
    """
    A definition for an endpoint.

    """
    def __new__(cls, func=None, request_schema=None, response_schema=None, header_func=None, response_formats=None):
        """
        Define an API endpoint.

        Defines the behavior of an API endpoint in conjunction with a `Namespace` and an `Operation`.

        Supports a callbable `func`, request and response (marshmallow) schemas, a header-modifying function, and
        optional response formats.

        The callable `func` should accept `**kwargs` and return a marshmallow-compatible object or dictionary.

        The callable `header_func` (if any) should accept a `headers` dictionary and the return value from the
        callable `func`.

        :param func: a function to process request data and return response data
        :param request_schema: a marshmallow schema to decode/validate request data
        :param response_schema: a marshmallow schema to encode response data
        :param header_func: a header-modifying function
        :param response_formats: an optional list of support response formats

        """
        return tuple.__new__(
            EndpointDefinition,
            (func, request_schema, response_schema, header_func, response_formats),
        )

    @property
    def func(self):
        return self[0]

    @property
    def request_schema(self):
        return self[1]

    @property
    def response_schema(self):
        return self[2]

    @property
    def header_func(self):
        return self[3] or (lambda headers, response_data: headers)

    @property
    def response_formats(self):
        return self[4] or []


class Convention:
    """
    A convention is a recipe for applying Flask-compatible functions to a namespace.

    """
    def __init__(self, graph):
        self.graph = graph
        self._registered_routes = set()

    def configure(self, ns, mappings=None, **kwargs):
        """
        Apply mappings to a namespace.

        """
        if mappings is None:
            mappings = dict()
        mappings.update(kwargs)

        for operation, definition in mappings.items():
            try:
                configure_func = self._find_func(operation)
            except AttributeError:
                pass
            else:
                configure_func(ns, self._make_definition(definition))

    def add_route(self, path, operation, ns):
        route = (operation.value.method, path)

        if route in self._registered_routes:
            raise RouteAlreadyRegisteredException(route)

        self._registered_routes.add(route)

        return self.graph.route(path, operation, ns)

    def _find_func(self, operation):
        """
        Find the function to use to configure the given operation.

        The input might be an `Operation` enum or a string.

        """
        if isinstance(operation, Operation):
            operation_name = operation.name.lower()
        else:
            operation_name = operation.lower()

        return getattr(self, "configure_{}".format(operation_name))

    def negotiate_response_content(self, allowed_response_formats):
        response_format = find_response_format(allowed_response_formats)
        if response_format is None:
            raise NotAcceptable()
        return response_format

    def _make_definition(self, definition):
        """
        Generate a definition.

        The input might already be a `EndpointDefinition` or it might be a tuple.

        """
        if not definition:
            return EndpointDefinition()
        if isinstance(definition, EndpointDefinition):
            return definition
        elif len(definition) == 1:
            return EndpointDefinition(
                func=definition[0],
            )
        elif len(definition) == 2:
            return EndpointDefinition(
                func=definition[0],
                response_schema=definition[1],
            )
        elif len(definition) == 3:
            return EndpointDefinition(
                func=definition[0],
                request_schema=definition[1],
                response_schema=definition[2],
            )
        elif len(definition) == 4:
            return EndpointDefinition(
                func=definition[0],
                request_schema=definition[1],
                response_schema=definition[2],
                header_func=definition[3],
            )



from functools import lru_cache
from pkg_resources import iter_entry_points
from typing import (
    Any,
    List,
    Mapping,
    Type,
)

from marshmallow.fields import Field

from microcosm_flask.swagger.parameters.base import ParameterBuilder
from microcosm_flask.swagger.parameters.default import DefaultParameterBuilder


ENTRY_POINT = "microcosm_flask.swagger.parameters"


class Parameters:
    """
    Plugin-aware swagger parameter builder.

    Discovers builder subclasses via the `microcosm_flask.swagger.parameters` entry point
    and delegates to the first compatible implementation.

    """
    def build(self, field: Field) -> Mapping[str, Any]:
        """
        Build a swagger parameter from a marshmallow field.

        """
        builder_types = self.builder_types() + [
            # put default last
            self.default_builder_type()
        ]

        builders: List[ParameterBuilder] = [
            builder_type(
                build_parameter=self.build,  # type: ignore
            )
            for builder_type in builder_types
        ]

        builder = next(
            builder
            for builder in builders
            if builder.supports_field(field)
        )

        return builder.build(field)

    @classmethod
    # NB: entry point lookups can be slow; memoize
    @lru_cache()
    def builder_types(cls) -> List[Type[ParameterBuilder]]:
        """
        Define the available builder types.

        """
        return [
            entry_point.load()
            for entry_point in iter_entry_points(ENTRY_POINT)
        ]

    @classmethod
    def default_builder_type(cls) -> Type[ParameterBuilder]:
        """
        Define the default builder type.

        """
        return DefaultParameterBuilder



from copy import deepcopy
from dataclasses import dataclass
from typing import List, Optional

from inflection import camelize, underscore
from marshmallow import Schema

from microcosm_flask.naming import name_for
from microcosm_flask.swagger.naming import type_name


@dataclass
class SelectedField:
    name: str
    required: Optional[bool] = None


def get_fields_from_schema(schema_cls: Schema, selected_fields: List[SelectedField]):
    """
    Select fields definitions from a schema

    By default, the selected fields will keep the same value for `required` as the source,
    but this can be overriden by passing the right value for `required` to the `SelectedField` instance
    passed in to this method.

    """
    associated_fields = {}
    for selected_field in selected_fields:
        if isinstance(selected_field, str):
            selected_field = SelectedField(selected_field)
        schema_field = deepcopy(schema_cls._declared_fields[selected_field.name])
        if selected_field.required is not None:
            schema_field.required = selected_field.required
        associated_fields[selected_field.name] = schema_field

    return associated_fields


def associated_schemas_attr_name(schema_cls):
    return f"_associated_schemas_{underscore(schema_cls.__name__)}"


def associated_schema_name(schema_cls, name_suffix):
    return f"{type_name(name_for(schema_cls))}{camelize(name_suffix)}Schema"


def get_associated_schema(schema_cls, name_suffix):
    associated_schemas = getattr(schema_cls, associated_schemas_attr_name(schema_cls), {})
    if name_suffix in associated_schemas:
        return associated_schemas[name_suffix]
    raise KeyError(f"Schema {schema_cls} does not have an associated schema with suffix {name_suffix}")


def set_associated_schema(target_cls, name_suffix, associated_schema):
    """
    Add a derived schema as an attribute to a schema class, using conventions
    that will result in the associated schema being exposed in the OpenAPI definition of the service.

    """
    # Use the class name in the attribute name to avoid sharing with children classes
    attr_name = associated_schemas_attr_name(target_cls)
    try:
        associated_schemas = getattr(target_cls, attr_name)
        if name_suffix in associated_schemas:
            raise ValueError(f"Schema {target_cls} already has an associated schema for suffix {name_suffix}")
        associated_schemas[name_suffix] = associated_schema
    except AttributeError:
        setattr(
            target_cls,
            attr_name,
            {name_suffix: associated_schema},
        )


def build_associated_schema(schema_cls, name_suffix, inherits_from, associated_fields):
    """
    Build a schema class, whose name is based on an existing schema.

    """
    return type(
        associated_schema_name(schema_cls, name_suffix),
        inherits_from,
        associated_fields,
    )


def add_associated_schema(name_suffix, selected_fields=(), inherits_from=(Schema,)):
    """
    Derive a schema as a subset of fields from the schema class being decorated,
    and add that derived schema as an attribute on the decorated schema.

    This allows us to expose the derived schema in the swagger definition.

    :name_suffix: Suffix that will added to the decorator schema's definition name to give the
                  name of the exposed definition
    :selected_fields: List of `SelectedField` instances. As a shorthand a list of strings can also
                      be passed, which will make all sected field required

    """
    def decorator(schema_cls):
        associated_fields = get_fields_from_schema(schema_cls, selected_fields)

        associated_schema = build_associated_schema(
            schema_cls,
            name_suffix,
            inherits_from,
            associated_fields,
        )

        set_associated_schema(schema_cls, name_suffix, associated_schema)
        return schema_cls

    return decorator



from typing import Optional

from marshmallow.fields import Decimal, Field

from microcosm_flask.swagger.parameters.base import ParameterBuilder


class NumericParameterBuilder(ParameterBuilder):
    """
    Build a string-formatted numeric parameter.

    """
    def supports_field(self, field: Field) -> bool:
        return bool(getattr(field, "as_string", None))

    def parse_format(self, field: Field) -> Optional[str]:
        if isinstance(field, Decimal):
            return "decimal"
        return None

    def parse_type(self, field: Field) -> str:
        return "string"



from marshmallow.fields import Field

from microcosm_flask.fields import TimestampField
from microcosm_flask.swagger.parameters.base import ParameterBuilder


class TimestampParameterBuilder(ParameterBuilder):
    """
    Build a timestamp parameter.

    """
    def supports_field(self, field: Field) -> bool:
        return isinstance(field, TimestampField)

    def parse_format(self, field: Field) -> str:
        if isinstance(field, TimestampField) and field.use_isoformat:
            return "date-time"
        return "timestamp"

    def parse_type(self, field: Field) -> str:
        if isinstance(field, TimestampField) and field.use_isoformat:
            return "string"
        return "float"



"""
Basic Auth support.

Usage:

    @graph.app.route("/path")
    @graph.basic_auth.required
    def must_be_authorized():
        pass


"""
from base64 import b64encode

from flask_basicauth import BasicAuth
from microcosm.api import defaults
from werkzeug.exceptions import Unauthorized

from microcosm_flask.conventions.encoding import with_headers


def encode_basic_auth(username, password):
    """
    Encode basic auth credentials.

    """
    return "Basic {}".format(
        b64encode(
            "{}:{}".format(
                username,
                password,
            ).encode("utf-8")
        ).decode("utf-8")
    )


class ConfigBasicAuth(BasicAuth):
    """
    Basic auth decorator that pulls credentials from static configuration.

    This decorator is sufficient for internal service access control, but should
    not be used for anything truly sensitive.

    """

    def __init__(self, app, credentials):
        super(ConfigBasicAuth, self).__init__(app)
        self.credentials = credentials

    def check_credentials(self, username, password):
        """
        Override credential checking to use configured credentials.

        """
        return password is not None and self.credentials.get(username, None) == password

    def challenge(self):
        """
        Override challenge to raise an exception that will trigger regular error handling.

        """
        response = super(ConfigBasicAuth, self).challenge()
        raise with_headers(Unauthorized(), response.headers)


@defaults(
    credentials=dict(
        # set a default configuration but don't merge it if other config is set
        __merge__=False,
        default="secret",
    ),
)
def configure_basic_auth_decorator(graph):
    """
    Configure a basic auth decorator.

    """
    # use the metadata name if no realm is defined
    graph.config.setdefault("BASIC_AUTH_REALM", graph.metadata.name)
    return ConfigBasicAuth(
        app=graph.flask,
        # wrap in dict to allow lists of items as well as dictionaries
        credentials=dict(graph.config.basic_auth.credentials),
    )



"""
A timestaamp-valued field.

Supports both UNIX epoch (preferred) and ISO 8601 (discouraged) encodings.

"""
from datetime import datetime

from dateutil import parser
from marshmallow.fields import Field, ValidationError


class TimestampField(Field):
    """
    Timestamp valued field, as either a unix timestamp (default) or isoformat string.

    """
    EPOCH = datetime(1970, 1, 1)

    def __init__(self, use_isoformat=False, *args, **kwargs):
        self.use_isoformat = use_isoformat
        super(TimestampField, self).__init__(*args, **kwargs)

    def _serialize(self, value, attr, obj, **kwargs):
        """
        Serialize value as a timestamp, either as a Unix timestamp (in float second) or a UTC isoformat string.

        """
        if value is None:
            return None

        if self.use_isoformat:
            return datetime.utcfromtimestamp(value).isoformat()
        else:
            return value

    def _deserialize(self, value, attr, obj, **kwargs):
        """
        Deserialize value as a Unix timestamp (in float seconds).

        Handle both numeric and UTC isoformat strings.

        """
        if value is None:
            return None

        try:
            return float(value)
        except ValueError:
            parsed = parser.parse(value)
            if parsed.tzinfo:
                if parsed.utcoffset().total_seconds():
                    raise ValidationError("Timestamps must be defined in UTC")
                parsed = parsed.replace(tzinfo=None)
            return (parsed - TimestampField.EPOCH).total_seconds()
